package ch.hslu.edu.enapp.webshop.boundary;

import java.io.Serializable;
import java.io.StringWriter;
import java.math.BigDecimal;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import javax.annotation.Resource;
import javax.ejb.EJB;
import javax.ejb.Local;
import javax.ejb.Remote;
import javax.ejb.Stateless;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.inject.Inject;
import javax.jms.Connection;
import javax.jms.ConnectionFactory;
import javax.jms.JMSException;
import javax.jms.Message;
import javax.jms.MessageProducer;
import javax.jms.Queue;
import javax.jms.Session;
import javax.persistence.EntityManager;
import javax.persistence.PersistenceContext;
import javax.transaction.UserTransaction;
import javax.xml.bind.JAXBContext;
import javax.xml.bind.Marshaller;

import org.apache.wink.client.ClientResponse;
import org.apache.wink.client.RestClient;

import ch.hslu.edu.enapp.webshop.boundary.dto.Ncresponse;
import ch.hslu.edu.enapp.webshop.common.CustomerServiceRemote;
import ch.hslu.edu.enapp.webshop.common.PurchaseServiceLocal;
import ch.hslu.edu.enapp.webshop.common.PurchaseServiceRemote;
import ch.hslu.edu.enapp.webshop.common.dto.CustomerDTO;
import ch.hslu.edu.enapp.webshop.common.dto.ProductDTO;
import ch.hslu.edu.enapp.webshop.common.dto.PurchaseDTO;
import ch.hslu.edu.enapp.webshop.common.dto.PurchaseitemDTO;
import ch.hslu.edu.enapp.webshop.control.ProductManager;
import ch.hslu.edu.enapp.webshop.control.PurchaseManager;
import ch.hslu.edu.enapp.webshop.control.PurchaseitemManager;
import ch.hslu.edu.enapp.webshop.enappdeamon.Customer;
import ch.hslu.edu.enapp.webshop.enappdeamon.Line;
import ch.hslu.edu.enapp.webshop.enappdeamon.PurchaseMessage;
import ch.hslu.edu.enapp.webshop.enappdeamon.SalesOrder;
import ch.hslu.edu.enapp.webshop.entity.Purchase;
import ch.hslu.edu.enapp.webshop.entity.Purchaseitem;

/**
 * Session Bean implementation class PurchaseService
 */
@Stateless
@Local(PurchaseServiceLocal.class)
@Remote(PurchaseServiceRemote.class)
@TransactionManagement(TransactionManagementType.BEAN)
public class PurchaseService implements PurchaseServiceRemote, PurchaseServiceLocal, Serializable {

	private static final long serialVersionUID = 1L;

	@Resource(name = "EnappQueue")
	Queue enappQueue;

	@Resource(name = "EnappQueueConnectionFactory")
	ConnectionFactory enappQueueConnectionFactory;

	@PersistenceContext
	EntityManager em;
	@Resource
	UserTransaction tx;

	@Inject
	private PurchaseManager purchaseManager;
	@Inject
	private PurchaseitemManager purchaseitemManager;
	@EJB
	private CustomerServiceRemote customerService;
	@Inject
	private ProductManager productManager;
	@Inject
	private PaymentService paymentService;

	/**
	 * Default constructor.
	 */
	public PurchaseService() {
		// empty
	}

	@Override
	public int execOrder(final Map<ProductDTO, Integer> shoppingCart) {
		final CustomerDTO customer = customerService.getCustomer();
		// System.out.println("execOrder User: " + customer.getUsername() + " (Id:" + customer.getId() + ")");

		// persist Order
		final int orderId = persistOrder(shoppingCart, customer);

		// exec Postfinance payment
		int ncStatus = -1;
		if (orderId > 0) {
			ncStatus = execPayment(orderId, shoppingCart, customer);
		}

		if (ncStatus == 0) {
			// send PurchaseMessage to EnappDaemon
			sendPurchase(orderId);
		}

		// return "/secured/orderdetail.xhtml?faces-redirect=true&order=" + purchase.getId(); // redirect to index.xhtml page
		return orderId;
	}

	private void sendPurchase(final int purchaseId) {
		final PurchaseDTO purchase = getPurchaseById(purchaseId, false);
		final String content = marshalMessage(getPurchaseMessage(purchase));

		System.out.println("CorrelationId: " + purchase.getCorrelationId());
		System.out.println("PurchaseMessage:\n" + content);

		sendJMSMessageToEnappQueue(purchase.getCorrelationId(), content);
	}

	private int execPayment(final int orderId, final Map<ProductDTO, Integer> shoppingCart, final CustomerDTO customer) {
		// Calc Order Amount
		double orderAmount = 0.0;
		final List<ProductDTO> products = new ArrayList<ProductDTO>();
		products.addAll(shoppingCart.keySet());
		for (final ProductDTO product : products) {
			orderAmount += shoppingCart.get(product) * product.getUnitprice().doubleValue();
		}

		// Exec Postfinance Payment!
		// generate special orderId for postfinance to avoid "already processed"...
		String postfinanceOrderId;
		final DateFormat df = new SimpleDateFormat("yyyyMMddHHmmss");
		final Date today = Calendar.getInstance().getTime();
		postfinanceOrderId = df.format(today) + "00" + orderId;

		final Ncresponse postfinanceResponse = paymentService.execPaymentRequest(postfinanceOrderId, orderAmount, customer);
		System.out.println("### OrderId:     " + orderId);
		System.out.println("### PfOrderId:   " + postfinanceOrderId);
		System.out.println("### Amount:      " + orderAmount);
		System.out.println("### PAYID:       " + postfinanceResponse.getPAYID());
		System.out.println("### NCSTATUS:    " + postfinanceResponse.getNCSTATUS());
		System.out.println("### NCERROR:     " + postfinanceResponse.getNCERROR());
		System.out.println("### NCERRORPLUS: " + postfinanceResponse.getNCERRORPLUS());

		// Update Purchase
		final Purchase order = purchaseManager.getPurchaseById(orderId).get(0);
		order.setPayid(postfinanceResponse.getPAYID());
		order.setNcstatus(postfinanceResponse.getNCSTATUS());
		order.setNcerror(postfinanceResponse.getNCERROR());
		order.setNcerrorplus(postfinanceResponse.getNCERRORPLUS());
		order.setTotalPrice(orderAmount);
		try {
			// System.out.println("Update Order...");
			if (postfinanceResponse.getNCERROR() == 0) {
				order.setStatus("paid");
			} else {
				order.setStatus("payment error: " + postfinanceResponse.getNCERRORPLUS());
			}
			tx.begin();
			em.merge(order);
			em.flush();
			tx.commit();
		} catch (final Exception e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

		return postfinanceResponse.getNCSTATUS();
	}

	private int persistOrder(final Map<ProductDTO, Integer> shoppingCart, final CustomerDTO customer) {

		final List<ProductDTO> products = new ArrayList<ProductDTO>();
		products.addAll(shoppingCart.keySet());

		// for (final ProductDTO product : products) {
		// System.out.println(shoppingCart.get(product) + "x " + product);
		// }

		final Purchase purchase = new Purchase();
		purchase.setCustomerid(customer.getId());
		purchase.setDatetime(new Timestamp(System.currentTimeMillis()));
		purchase.setStatus("ordered");
		purchase.setCorrelationId(UUID.randomUUID().toString());
		System.out.println("New CorrelationId: " + purchase.getCorrelationId());

		Purchaseitem item;
		int quantity;
		BigDecimal quantityDb, amountDb;
		try {
			try {
				tx.begin();

				em.persist(purchase);
				em.flush();
				em.merge(purchase);

				for (final ProductDTO product : products) {
					quantity = shoppingCart.get(product);
					quantityDb = new BigDecimal(quantity);
					amountDb = quantityDb.multiply(product.getUnitprice());

					item = new Purchaseitem();
					item.setPurchaseid(purchase.getId());
					item.setProductid(product.getMsDynNAVItemNo());
					item.setDescription(product.getName());
					item.setUnitprice(product.getUnitprice());
					item.setQuantity(quantityDb);
					item.setLineamount(amountDb);

					em.persist(item);
				}

				// System.out.println("PurchaseId: " + purchase.getId());
			} catch (final Exception e) {
				// rollback transaction
				tx.rollback();

				// TODO: handle exception
				System.err.println("Persist Order Failed! (1)");
				e.printStackTrace();

				// return "/error.xhtml?faces-redirect=true"; // redirect to error.xhtml page
				return -1;
			} finally {
				em.flush();
				tx.commit();
			}
		} catch (final Exception e) {
			// transaction failed
			System.err.println("Persist Order Failed! (2)");
			e.printStackTrace();

			// TODO: handle exception
			// return "/error.xhtml?faces-redirect=true"; // redirect to error.xhtml page
			return -2;
		}

		// return "/secured/orderdetail.xhtml?faces-redirect=true&order=" + purchase.getId(); // redirect to index.xhtml page
		return purchase.getId();
	}

	@Override
	public List<PurchaseDTO> getPurchaseHistory() {
		final CustomerDTO customer = customerService.getCustomer();
		if ((customer == null) || (customer.getId() == 0)) {
			return new ArrayList<PurchaseDTO>();
		}

		final ArrayList<PurchaseDTO> historyList = new ArrayList<PurchaseDTO>();
		final List<Purchase> purchases = purchaseManager.getPurchaseByCustomerid(customer.getId());
		for (final Purchase purchase : purchases) {
			historyList.add(PurchaseConverter.createDTOFromEntity(purchase));
		}
		return historyList;
	}

	@Override
	public PurchaseDTO getPurchaseById(final int purchaseId) {
		return getPurchaseById(purchaseId, false);
	}

	@Override
	public PurchaseDTO getPurchaseById(final int purchaseId, final boolean reloadStatus) {
		final CustomerDTO customer = customerService.getCustomer();
		if ((customer == null) || (customer.getId() == 0)) {
			return new PurchaseDTO();
		}

		// Update PurchaseStatus
		if (reloadStatus) {
			final String res = updateSalesOrderStatus(purchaseId);
			System.out.println("$$$ res: " + res);
		}

		final List<Purchase> purchaseList = purchaseManager.getPurchaseById(purchaseId);
		if (!purchaseList.isEmpty()) {
			if ((customerService.isAdmin(customer)) || (purchaseList.get(0).getCustomerid() == customer.getId())) {
				// return this purchase only if its from the current logged on users
				return PurchaseConverter.createDTOFromEntity(purchaseList.get(0));
			}
		}

		// if ID not found (or not submitted) select newest purchase from this customer
		final List<PurchaseDTO> purchaseDTOList = getPurchaseHistory();
		if (!purchaseDTOList.isEmpty()) {
			if (purchaseDTOList.get(purchaseDTOList.size() - 1).getCustomerid() == customer.getId()) {
				// return this purchase only if its from the current logged on users
				return purchaseDTOList.get(purchaseDTOList.size() - 1);
			}
		}

		// return empty purchaseDTO
		return new PurchaseDTO();

	}

	@Override
	public List<PurchaseitemDTO> getPurchaseitemsByPurchaseId(final int purchaseId) {
		final CustomerDTO customer = customerService.getCustomer();
		if ((customer == null) || (customer.getId() == 0)) {
			return new ArrayList<PurchaseitemDTO>();
		}

		final PurchaseDTO purchase = getPurchaseById(purchaseId, false);
		if (purchase.getId() == 0) {
			return new ArrayList<PurchaseitemDTO>();
		}

		// final Product product;
		// final List<Product> productList;
		PurchaseitemDTO purchaseitemDTO;
		final List<PurchaseitemDTO> purchaseitemListDTO = new ArrayList<PurchaseitemDTO>();

		final List<Purchaseitem> purchaseitemList = purchaseitemManager.getPurchaseitemByPurchaseid(purchase.getId());
		for (final Purchaseitem purchaseitem : purchaseitemList) {
			purchaseitemDTO = PurchaseitemConverter.createDTOFromEntity(purchaseitem);

			// add Product Info (Milestone 1)
			/*
			 * productList = productManager.getProductById(purchaseitem.getProductid()); if (!productList.isEmpty()) { product = productList.get(0); // Product
			 * still exists - use this name and Description purchaseitemDTO.setName(product.getName());
			 * purchaseitemDTO.setDescription(product.getDescription()); } else { // use order item Description
			 * purchaseitemDTO.setName(purchaseitemDTO.getDescription()); }
			 */

			purchaseitemListDTO.add(purchaseitemDTO);
		}

		return purchaseitemListDTO;
	}

	@Override
	public List<PurchaseDTO> getAllPurchaseHistory() {
		final ArrayList<PurchaseDTO> historyList = new ArrayList<PurchaseDTO>();
		final List<Purchase> purchases = purchaseManager.getPurchase();
		for (final Purchase purchase : purchases) {
			historyList.add(PurchaseConverter.createDTOFromEntity(purchase));
		}
		return historyList;
	}

	private void sendJMSMessageToEnappQueue(final String correlationId, final String content) {
		Connection connection = null;
		Session session = null;
		try {
			connection = enappQueueConnectionFactory.createConnection();
			session = connection.createSession(false, Session.AUTO_ACKNOWLEDGE);
			final MessageProducer messageProducer = session.createProducer(enappQueue);
			final Message message = session.createTextMessage(content);
			message.setStringProperty("MessageFormat", "Version 1.5");
			message.setJMSCorrelationID(correlationId);
			message.setJMSReplyTo(null); // we don't use any reply queue :(
			messageProducer.send(message); // send message :)
		} catch (final Exception e) {
			e.printStackTrace();
		} finally {
			if (session != null) {
				try {
					session.close();
				} catch (final JMSException e) {
					System.out.println("Cannot close session: " + e);
				}
			}
			if (connection != null) {
				try {
					connection.close();
				} catch (final JMSException e) {
					System.out.println("Cannot close connection: " + e);
				}
			}
		}
	}

	@Override
	public String updateSalesOrderStatus(final int purchaseId) {
		String result = "error";

		Purchase purchase = null;
		final List<Purchase> purchaseList = purchaseManager.getPurchaseById(purchaseId);
		if (!purchaseList.isEmpty()) {
			purchase = purchaseList.get(0);
		} else {
			return result; // error
		}

		SalesOrder salesOrderStatus = new SalesOrder();
		salesOrderStatus.setOrderStatus(result);
		try {
			// Query NAV (EnappDaemon) OrderStatus
			salesOrderStatus = getOrderState(purchase.getCorrelationId());
			result = salesOrderStatus.getOrderStatus();
		} catch (final Exception e) {
			return result; // error
		}

		try {
			if (!result.equals("error")) {
				purchase.setNAVOrderId(salesOrderStatus.getExternalOrderId());
				purchase.setStatus(result);
				tx.begin();
				em.merge(purchase);
				em.flush();
				tx.commit();

				// Update Customer / NAV CustomerId
				final int customerId = purchase.getCustomerid();
				final CustomerDTO customerDTO = customerService.getCustomerById(customerId);
				final ch.hslu.edu.enapp.webshop.entity.Customer customer = CustomerConverter.createEntityFromDTO(customerDTO);
				customer.setNAVCustomer(salesOrderStatus.getExternalCustomerId());

				tx.begin();
				em.merge(customer);
				em.flush();
				tx.commit();
			}
		} catch (final Exception e) {
			return "error"; // error
		}

		return result;
	}

	private PurchaseMessage getPurchaseMessage(final PurchaseDTO aPurchase) {
		final PurchaseMessage message = new PurchaseMessage();
		message.setPurchaseId(aPurchase.getId());
		message.setPaymentId(aPurchase.getPayid());
		message.setStudent("tdrohrer");
		message.setTotalAmount(aPurchase.getTotalPrice());
		message.setDate(aPurchase.getDatetime());
		message.setCustomer(getCustomer(aPurchase.getCustomerid()));
		message.setLines(getLines(aPurchase.getId()));
		return message;
	}

	private Customer getCustomer(final int customerId) {
		final CustomerDTO entity = customerService.getCustomerById(customerId);
		final Customer customer = new Customer();
		customer.setExternalCustomerId(entity.getNAVCustomer());
		customer.setFullName(entity.getName());
		// customer.setAddress(entity.getAddress());
		customer.setAddress("Technikumstrasse");
		customer.setPostCode("6048");
		customer.setCity("Horw");
		customer.setUsername(entity.getUsername());
		return customer;
	}

	private Collection<Line> getLines(final int purchaseId) {
		final Collection<Line> lines = new ArrayList<Line>();
		for (final PurchaseitemDTO entity : getPurchaseitemsByPurchaseId(purchaseId)) {
			final Line line = new Line();
			line.setProductId(entity.getProductid());
			line.setDescription(entity.getDescription());
			line.setQuantity(entity.getQuantity().longValue());
			line.setAmount(entity.getLineamount().longValue());
			lines.add(line);
		}
		return lines;
	}

	private String marshalMessage(final PurchaseMessage message) {
		String textMessage = null;
		try {
			final JAXBContext context = JAXBContext.newInstance(PurchaseMessage.class);
			final StringWriter writer = new StringWriter();
			final Marshaller marshaller = context.createMarshaller();
			marshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
			marshaller.marshal(message, writer);
			textMessage = writer.toString();
		} catch (final Exception e) {
			System.out.println("Error MarshalMessage: " + e.getMessage());
		}
		return textMessage;
	}

	public SalesOrder getOrderState(final String correlationId) {
		// private static final String URL = "http://enapp-daemons.el.eee.intern:9080/EnappDaemonWeb/rest/salesorder/corr/%s";
		final String requestURI = "http://10.177.1.5:9080/EnappDaemonWeb/rest/salesorder/corr/" + correlationId;

		System.out.println("REST SalesOrderStatus URL: " + requestURI);
		final RestClient client = new RestClient();
		final org.apache.wink.client.Resource webResource = client.resource(requestURI);

		SalesOrder salesOrderStatus = new SalesOrder();
		salesOrderStatus.setOrderStatus("error");
		try {
			final ClientResponse response = webResource.get();
			salesOrderStatus = response.getEntity(SalesOrder.class);
		} catch (final Exception e) {
			// EnAPP Daemon not yet up-to-date... ignore :)
		}

		// System.out.println("# MSDynNAV SalesOrderNo: " + salesOrderStatus.getExternalOrderId());
		// System.out.println("# MSDynNAV CustomerNo:   " + salesOrderStatus.getExternalCustomerId());
		// System.out.println("# MSDynNAV OrderStatus:  " + salesOrderStatus.getOrderStatus());

		return salesOrderStatus;
	}

}
