package pagecode;

import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.Collections;
import java.util.Iterator;
import java.util.Map;

import javax.faces.component.NamingContainer;
import javax.faces.component.UIComponent;
import javax.faces.component.UIComponentBase;
import javax.faces.component.UIViewRoot;
import javax.faces.context.FacesContext;

/**
 * Provides a common base class for all generated Page Code files.
 */
@SuppressWarnings({ "rawtypes", "unchecked" })
public abstract class PageCodeBase {

	public PageCodeBase() {
	}

	protected void gotoPage(final String pageName) {
		if (pageName != null) {
			final FacesContext context = getFacesContext();
			final UIViewRoot newView = context.getApplication().getViewHandler().createView(context, pageName);
			context.setViewRoot(newView);
			context.renderResponse();
		}
	}

	/**
	 * <p>
	 * Return the {@link UIComponent} (if any) with the specified <code>id</code>, searching recursively starting at the specified <code>base</code>, and
	 * examining the base component itself, followed by examining all the base component's facets and children. Unlike findComponent method of
	 * {@link UIComponentBase}, which skips recursive scan each time it finds a {@link NamingContainer}, this method examines all components, regardless of
	 * their namespace (assuming IDs are unique).
	 * 
	 * @param base
	 *            Base {@link UIComponent} from which to search
	 * @param id
	 *            Component identifier to be matched
	 */
	public static UIComponent findComponent(final UIComponent base, final String id) {

		// Is the "base" component itself the match we are looking for?
		if (id.equals(base.getId())) {
			return base;
		}

		// Search through our facets and children
		UIComponent kid = null;
		UIComponent result = null;
		final Iterator kids = base.getFacetsAndChildren();
		while (kids.hasNext() && (result == null)) {
			kid = (UIComponent) kids.next();
			if (id.equals(kid.getId())) {
				result = kid;
				break;
			}
			result = findComponent(kid, id);
			if (result != null) {
				break;
			}
		}
		return result;
	}

	public static UIComponent findComponentInRoot(final String id) {
		UIComponent ret = null;

		final FacesContext context = FacesContext.getCurrentInstance();
		if (context != null) {
			final UIComponent root = context.getViewRoot();
			ret = findComponent(root, id);
		}

		return ret;
	}

	/**
	 * Place an Object on the tree's attribute map
	 * 
	 * @param key
	 * @param value
	 */
	protected void putTreeAttribute(final String key, final Object value) {
		getFacesContext().getViewRoot().getAttributes().put(key, value);
	}

	/**
	 * Retrieve an Object from the tree's attribute map
	 * 
	 * @param key
	 * @return
	 */
	protected Object getTreeAttribute(final String key) {
		return getFacesContext().getViewRoot().getAttributes().get(key);
	}

	/**
	 * Return the result of the resolved expression
	 * 
	 * @param expression
	 * @return
	 */
	@SuppressWarnings("deprecation")
	protected Object resolveExpression(final String expression) {
		Object value = null;
		if ((expression.indexOf("#{") != -1) && (expression.indexOf("#{") < expression.indexOf('}'))) {
			value = getFacesContext().getApplication().createValueBinding(expression).getValue(getFacesContext());
		} else {
			value = expression;
		}
		return value;
	}

	/**
	 * Return the managed bean with the given name
	 * 
	 * @param mgdBeanName
	 *            the name of the managed bean to retrieve
	 * @return
	 */
	protected Object getManagedBean(final String mgdBeanName) {
		final String expression = "#{" + mgdBeanName + "}";
		return resolveExpression(expression);
	}

	/**
	 * Resolve all parameters passed in via the argNames/argValues array pair, and add them to the provided paramMap. If a parameter can not be resolved, then
	 * it will attempt to be retrieved from a cachemap stored using the cacheMapKey
	 * 
	 * @param paramMap
	 * @param argNames
	 * @param argValues
	 * @param cacheMapKey
	 */
	protected void resolveParams(final Map paramMap, final String[] argNames, final String[] argValues, final String cacheMapKey) {

		final Object rawCache = getTreeAttribute(cacheMapKey);
		Map cache = Collections.EMPTY_MAP;
		if (rawCache instanceof Map) {
			cache = (Map) rawCache;
		}
		for (int i = 0; i < argNames.length; i++) {
			Object result = resolveExpression(argValues[i]);
			if (result == null) {
				result = cache.get(argNames[i]);
			}
			paramMap.put(argNames[i], result);
		}
		putTreeAttribute(cacheMapKey, paramMap);
	}

	/**
	 * Returns a full system path for a file path given relative to the web project
	 */
	protected static String getRealPath(final String relPath) {
		String path = relPath;
		try {
			final URL url = FacesContext.getCurrentInstance().getExternalContext().getResource(relPath);
			if (url != null) {
				path = url.getPath();
			}
		} catch (final MalformedURLException e) {
			e.printStackTrace();
		}
		return path;
	}

	/**
	 * Returns an InputStream for a resource at the given path
	 */
	protected static InputStream getResourceInputStream(final String relPath) {
		return FacesContext.getCurrentInstance().getExternalContext().getResourceAsStream(relPath);
	}

	protected void logException(final Throwable throwable) {
		final StringWriter stringWriter = new StringWriter();
		final PrintWriter printWriter = new PrintWriter(stringWriter);
		throwable.printStackTrace(printWriter);
		log(stringWriter.toString());
	}

	protected void log(final String message) {
		System.out.println(message);
	}

	public Map getApplicationScope() {
		return getFacesContext().getExternalContext().getApplicationMap();
	}

	public FacesContext getFacesContext() {
		return FacesContext.getCurrentInstance();
	}

	public Map getRequestParam() {
		return getFacesContext().getExternalContext().getRequestParameterMap();
	}

	public Map getRequestScope() {
		return getFacesContext().getExternalContext().getRequestMap();
	}

	public Map getSessionScope() {
		return getFacesContext().getExternalContext().getSessionMap();
	}

}